/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _ALMUSVCU_H_
#define _ALMUSVCU_H_

#include <stdio.h>
#include <dirent.h>
#include <inttypes.h>
#include <sys/select.h>

#include "defs.h"
#include "apui_defs.h"

/*
#ifndef true
#define true 1
#define false 0
typedef int bool_t;
#endif
*/

/* FIXME: defines in apui_defs.h should be here */

/*
 * Struct for describing a filter program as stored as a file on disk. 
 */
typedef struct _prgid_t_ {
    unsigned char md5sum[16];
} prgid_t;

struct filterprg {
    /* unique identifier */
    prgid_t id;
    /* Path to the file containing the filter program */
    char path[PATH_MAX];
    /* type of filter program */
#define FILTERPRG_TYPE_EQUALISATION                    1
#define FILTERPRG_TYPE_CROSSTALK_CANCELLATION          2
#define FILTERPRG_TYPE_REVERB_TRADITIONAL              10
#define FILTERPRG_TYPE_REVERB_BFORMAT                  11
    int type;
#define FILTERPRG_SUBTYPE_BFORMAT_NORMALISED 1
    int subtype;
    /* sample format */
    int format;
    /* source layout */
    int source_layout; /* VCU_REVERB_SOURCE_LAYOUT_* */
    /* number of channels of the program */
    int n_channels;
    /* number of (decorrelated) copies of the impulse responses */
    int n_sets;
    /* for traditional reverb programs, direction of microphones */
    int n_receivers;
    double recv_azimuth[FILTERPRG_CHANNELS_MAX];
    double recv_elevation[FILTERPRG_CHANNELS_MAX];
    double recv_distance[FILTERPRG_CHANNELS_MAX];
    /* name of the filter program */
    char name[FILTERPRG_NAME_MAX];
    /* long name of the filter program */
    char longname[FILTERPRG_LONGNAME_MAX];
    /* the number of sample rates available */
    int n_rates;
    /* which 'n_rates' sample rates, in Hz */
    int rates[FILTERPRG_RATES_MAX];
    /* how much to scale in dB so the (possibly left out) direct sound
       component becomes 0 dB */
    double scale;    
    double predelay;
    /* length in seconds */
    double length;
};

struct vcu_config {
    char path[PATH_MAX];
#define VCU_CONFIG_NAME_SIZE 21    
    char name[VCU_CONFIG_NAME_SIZE];
    char static_config_path[PATH_MAX];
    char cwd[PATH_MAX];
#define VCU_CONFIG_ARGS_MAX 8
    int argc;
#define VCU_CONFIG_ARG_SIZE 128
    char argv[VCU_CONFIG_ARGS_MAX][VCU_CONFIG_ARG_SIZE];
};

#define VCU_PHASE_MINIMUM 1
#define VCU_PHASE_MAXIMUM 2
#define VCU_PHASE_LINEAR 3

struct vcu_eqpart {
#define VCU_EQPRG_TYPE_OP_FLAG 0x8000
#define VCU_EQPRG_TYPE_OP_CONVOLVE (1 | VCU_EQPRG_TYPE_OP_FLAG)
#define VCU_EQPRG_TYPE_OP_ADD      (2 | VCU_EQPRG_TYPE_OP_FLAG)
#define VCU_EQPRG_TYPE_OP_SUBTRACT (3 | VCU_EQPRG_TYPE_OP_FLAG)
#define VCU_EQPRG_TYPE_OP_MULTIPLY (4 | VCU_EQPRG_TYPE_OP_FLAG)   
#define VCU_EQPRG_TYPE_LOWPASS      1
#define VCU_EQPRG_TYPE_HIGHPASS     2
#define VCU_EQPRG_TYPE_BANDPASS     3
#define VCU_EQPRG_TYPE_BANDSTOP     4
#define VCU_EQPRG_TYPE_ALLPASS      5
#define VCU_EQPRG_TYPE_CUSTOM       6
#define VCU_EQPRG_TYPE_RANDOM_PHASE 7
#define VCU_EQPRG_TYPE_FFT_STATIC   8
#define VCU_EQPRG_TYPE_FFT_DYNAMIC  9
    int type;
    union {
        /* operations */
        struct {
            double amp;
            int sign;
        } multiply;

        /* filters */
	struct {
	    double cutoff;
	    double slope;
	    int phase;            
	} lowpass;
	struct {
	    double cutoff;
	    double slope;
	    int phase;
	} highpass;
	struct {
	    double cutoff[2];
	    double slope;
	    int phase;
	} bandpass;
	struct {
	    double cutoff[2];
	    double slope;
	    int phase;
	} bandstop;
	struct {
	    prgid_t id;
	} custom;
	struct {
	    double length;
	} random_phase;
        struct {
#define VCU_EQPRG_MAX_BANDS 32            
            int n_bands;
#define VCU_EQPRG_MIN_BAND_FREQUENCY 5
#define VCU_EQPRG_MAX_BAND_FREQUENCY 20000
            double band[VCU_EQPRG_MAX_BANDS];
            double amp[VCU_EQPRG_MAX_BANDS];
        } fft;
    } params;
};

union vcu_lexval {
    bool_t boolean;
    double real;
    char *string;
    char *field;
};

#define VCU_LEX_EOS     1 /* end of statement (;) */
#define VCU_LEX_COMMA   2 /* , */
#define VCU_LEX_SLASH   3 /* / */
#define VCU_LEX_LBRACE  4 /* { */
#define VCU_LEX_RBRACE  5 /* } */

#define VCU_LEXVAL_BOOLEAN 100
#define VCU_LEXVAL_REAL    101
#define VCU_LEXVAL_STRING  102
#define VCU_LEXVAL_FIELD   103

#define VCU_FUN_NAME "vcu_name"
#define VCU_FUN_PREINIT "vcu_preinit"
#define VCU_FUN_INIT "vcu_init"
#define VCU_FUN_SHARE "vcu_share"
#define VCU_FUN_SELECT_WRAPPER "vcu_select_wrapper"
#define VCU_FUN_FINISH "vcu_finish"
#define VCU_FUN_NEEDTERM "vcu_needterm"
#define VCU_FUN_LOADUSERCONF "vcu_loaduserconf"
#define VCU_FUN_SAVEUSERCONF "vcu_saveuserconf"
#define VCU_FUN_PROGRESS_INIT_PHASE "vcu_progress_init_phase"
#define VCU_FUN_PROGRESS_BENCHMARK "vcu_progress_benchmark"
#define VCU_FUN_PROGRESS_LOAD_PROGRAM "vcu_progress_load_program"
#define VCU_FUN_PROGRESS_GENERATE_EQ "vcu_progress_generate_eq"
#define VCU_FUN_PROGRESS_GENERATE_AMBISONIC "vcu_progress_generate_ambisonic"

struct vcu;

const char *
vcu_name(void);

bool_t
vcu_preinit(int (*get_config_token)(union vcu_lexval *lexval));

bool_t
vcu_init(const struct vcu *_vcu);

void
vcu_finish(void);

bool_t
vcu_needterm(void);

void *
vcu_share(void);

int
vcu_select_wrapper(int n,
                   fd_set *readfds,
                   fd_set *writefds,
                   fd_set *exceptfds,
                   struct timeval *timeval);

bool_t
vcu_loaduserconf(int (*get_config_token)(union vcu_lexval *lexval));

void
vcu_saveuserconf(void (*put_config_token)(int tok,
                                          const union vcu_lexval *lexval));

void
vcu_progress_init_phase(int phase);

void
vcu_progress_generate_ambisonic(int filter_index,
                                int n_filters);

void
vcu_progress_generate_eq(int eq_index,
                         int n_equalisers,
                         int error_code);

void
vcu_progress_benchmark(int sample_rate,
                       double length,
                       double delay,
                       double load);

void
vcu_progress_load_program(const struct filterprg *prg,
                          int channel_index,
                          int n_channels,
                          int from_rate,
                          int to_rate,
                          bool_t deriving);

struct ui_module_functions {
    const char *(*name)(void);
    bool_t (*preinit)(int (*get_config_token)(union vcu_lexval *lexval));
    bool_t (*init)(const struct vcu *_vcu);
    void (*finish)(void);
    bool_t (*needterm)(void);
    void *(*share)(void);
    int (*select_wrapper)(int n,
                          fd_set *readfds,
                          fd_set *writefds,
                          fd_set *exceptfds,
                          struct timeval *timeval);
    /* should set default values on failure, or if input param is NULL */
    bool_t (*loaduserconf)(int (*get_config_token)(union vcu_lexval *lexval));
    void (*saveuserconf)(void (*put_config_token)(int tok,
						  const union vcu_lexval
						  *lexval));
    struct ui_progress_callbacks {
#define VCU_INIT_PHASE_CONVOLVER_SUBSYSTEM 1
#define VCU_INIT_PHASE_LOAD_REVERB_PROGRAM 2
#define VCU_INIT_PHASE_LOAD_XTC_PROGRAM 3
#define VCU_INIT_PHASE_GENERATING_EQUALISERS 4
#define VCU_INIT_PHASE_GENERATING_AMBISONIC 5
#define VCU_INIT_PHASE_FINISHED 6
	void (*init_phase)(int phase);
	void (*benchmark)(int sample_rate,
			  double length,
			  double delay,
			  double load);
	void (*load_program)(const struct filterprg *prg,
			     int channel_index,
			     int n_channels,
			     int from_rate,
			     int to_rate,
                             bool_t deriving);
	void (*generate_eq)(int eq_index,
			    int n_equalisers,
			    int error_code);
        void (*generate_ambisonic)(int filter_index,
                                   int n_filters);
    } progress;
};

struct ui_module {
    char *name;
    void *shared_data;
};

/* FIXME: go through the use of the error codes */
#define VCU_ERR_UNKNOWN_ERROR               -1
#define VCU_ERR_INVALID_VALUE               -2
#define VCU_ERR_PREREQUISITES_NOT_FULFILLED -3
#define VCU_ERR_NOT_IMPLEMENTED             -4
#define VCU_ERR_NOT_SUPPORTED_BY_HARDWARE   -5
#define VCU_ERR_INCORRECT_STATE             -6
#define VCU_ERR_NO_SIGNAL                   -7
#define VCU_ERR_UNSUPPORTED_SAMPLE_RATE     -8
#define VCU_ERR_NO_MEMORY                   -9
#define VCU_ERR_LOAD_FAILED                 -10
#define VCU_ERR_RESAMPLING_FAILED           -11
#define VCU_ERR_WRITE_FAILED                -12
#define VCU_ERR_READ_FAILED                 -13
#define VCU_ERR_OBJECT_BUSY                 -14
#define VCU_ERR_OBJECT_TOO_LARGE            -15
#define VCU_ERR_INVALID_FORMAT              -16
#define VCU_ERR_CONVOLVER_FAILED_UNKNOWN    -17
#define VCU_ERR_CONVOLVER_FAILED_UNDERFLOW  -18
#define VCU_ERR_CONVOLVER_FAILED_NO_MEMORY  -19
#define VCU_ERR_CONVOLVER_FAILED_SIGNAL     -20
#define VCU_ERR_OBJECT_EXISTS               -21
#define VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX -22
#define VCU_ERR_TRUNCATION_NECESSARY        -23

/* FIXME: make these function calls? rates_get_count, or perhaps the other
   way around, make all ranges and counts to defines to simplify
   implementation */
#define VCU_VALID_RATES_COUNT 4
#define VCU_VALID_RATES { 44100, 48000, 88200, 96000 }

/* FIXME: refers to BFIR define */
#define VCU_BENCHMARK_MAX_ALTERNATIVES BFIR_BENCHMARK_N_BLOCK_SIZES

#define VCU_CH_AMBIOPOLE 0
#define VCU_CH_REVERB 1
#define VCU_CH_AMBISONIC 2
#define VCU_CH_PASSTHRU 3
#define VCU_CH_COUNT 4

#define VCU_CH_PHYSICAL 4
#define VCU_CH_WP_COUNT 5

#define	VCU_PROGRAM_TYPE_REVERB 1
#define	VCU_PROGRAM_TYPE_XTC    2
#define	VCU_PROGRAM_TYPE_EQ     3

/* Ambisonic decoder setting */
#define VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER  1
#define VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER 2
#define VCU_AMBISONIC_INPUT_FORMAT_UHJ          3

/* Reverb settings */
#define VCU_REVERB_RESP_MODE_FULL          1
#define VCU_REVERB_RESP_MODE_MUTE_DIRECT   2
#define VCU_REVERB_RESP_MODE_MUTE_EARLY    3
#define VCU_REVERB_RESP_MODE_STAGE_ADAPTED 4

#define VCU_REVERB_DECORR_MODE_ALL    1
#define VCU_REVERB_DECORR_MODE_NEEDED 2
#define VCU_REVERB_DECORR_MODE_NONE   3

#define VCU_REVERB_SOURCE_LAYOUT_CENTER 1
#define VCU_REVERB_SOURCE_LAYOUT_STEREO 2
#define VCU_REVERB_SOURCE_LAYOUT_LRC    3

/* In/out index for channels */
#define VCU_IN 0
#define VCU_OUT 1

/* Not applicable sysinfo value */
#define VCU_SYSINFO_NA -500

#define VCU_BENCHMARK_LIMIT_CPU 1
#define VCU_BENCHMARK_LIMIT_MEMORY 2
#define VCU_BENCHMARK_LIMIT_HARD 3

struct vcu_chgrp {
#define VCU_CHGRP_NAME_SIZE 8
    char name[VCU_CHGRP_NAME_SIZE];
    uint32_t channels[2][VCU_CH_COUNT][VCU_MAX_CHANNELS/32];
};

struct vcu_chmode {
#define VCU_CHMODE_NAME_SIZE 21
    char name[VCU_CHMODE_NAME_SIZE];
    uint32_t groups[VCU_MAX_CHANNEL_GROUPS/32];
    uint32_t eqmodes[VCU_MAX_EQ_MODES/32];
    int active_eq_mode;
};

struct vcu_eqmode {
#define VCU_EQMODE_NAME_SIZE 8    
    char name[VCU_EQMODE_NAME_SIZE];
    int prgs[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
};

struct vcu_eqprg {
#define VCU_EQPRG_NAME_SIZE 12
    char name[VCU_EQPRG_NAME_SIZE];
    struct vcu_eqpart *parts;
    int part_count;
};

struct vcu_eqamp {
    double in;
    double thru;
    double out;
    double mean;
    double min;
    double max;
};

struct vcu {
/****
 * vcu.runtime.*
 *
 * A set of functions used to start and stop the convolver engine, as well to
 * change its parameters in runtime. Note that the parameters can be adjusted
 * also when the convolver engine is not running, however most are for the user
 * most naturally accessed during runtime when direct feedback is available
 * (for example changing the master volume and hearing it change at the same
 * time).
 */
    struct /* runtime */ {
        /****
         * vcu.runtime.isruntime
         *
         * Returns true if the convolver engine is running, else false.
         */
	bool_t (*isruntime)(void);
#define VCU_PEAK_NEG_INFINITE -10000.0
	int (*start)(void (*peak_meter)(double peak),
		     void (*process_terminated)(int reason));
        int (*start_file_input)(void (*peak_meter)(double peak),
                                void (*process_terminated)(int reason),
                                int sample_format,
                                int channel_count,
                                bool_t loop,
                                const char path[]);
	void (*stop)(void);
        void (*buffer_commands)(bool_t do_buffer);
	struct { /* signal */
	    /****
	     * vcu.runtime.signal.get_input_sample_rate
	     *
	     * Get the current sample rate in Hz. Will return
	     * VCU_ERR_INVALID_VALUE if there is no valid input.
	     */
	    int (*get_input_sample_rate)(void);
	    void (*reset_peak_meter)(void);
	} signal;
	struct { /* volume */
	    void (*get_range)(double range[2]);
	    double (*get_master)(void);
	    int (*set_master)(double volume);
	    double (*get_group)(int idx);
	    int (*set_group)(int idx,
                             double volume);
	    bool_t (*get_mute)(int io,
                               int type, /* strictly per channel, w phys */
			       int idx);
	    int (*set_mute)(int io,
                            int type, /* strictly per channel, w phys */
			    int idx,
			    bool_t mute);            
	    bool_t (*get_group_mute)(int idx);
	    int (*set_group_mute)(int idx,
                                  bool_t mute);
            
	    bool_t (*get_master_temp_mute)(void);
	    int (*set_master_temp_mute)(bool_t mute);
            
	    double (*get_trim)(int io,
                               int type, /* strictly per channel, w phys */
			       int idx);
	    int (*set_trim)(int io,
                            int type, /* strictly per channel, w phys */
			    int idx,
			    double volume);
	} volume;
	struct { /* delay */
	    double (*get_max)(void);
	    double (*get_trim)(int io,
                               int type, /* strictly per channel, w phys */
			       int idx);
	    int (*set_trim)(int io,
                            int type, /* strictly per channel, w phys */
			    int idx,
			    double delay);
	} delay;
	struct { /* eq */
            const int *(*get_dynamic_equalisers)(int *count,
                                                 bool_t only_active);
            int (*modify_dynamic_equaliser)(int idx,
                                            const struct vcu_eqpart *eqpart);
            void (*get_dynamic_equaliser_volume_range)(double range[2]);
            double (*get_dynamic_equaliser_smallest_band)(void);

            int (*get_mode)(void);
            int (*set_mode)(int idx);
            bool_t (*get_mode_active)(int idx);
	} eq;
	struct { /* reverb */
	    prgid_t (*get_id)(void);
	    int (*set_id)(prgid_t id);
	    /****
	     * vcu.runtime.reverb.get_trim_volume_range
	     * vcu.runtime.reverb.get_trim_volume
	     * vcu.runtime.reverb.set_trim_volume
	     *
	     * This sets the overall trim volume for the currently selected
	     * reverb program. The purpose of this is to be able to adjust the
	     * volume of the stored reverb program if it is much different
	     * compared to other programs. The setting is stored in a reverb
	     * configuration file on disk.
	     *
	     * .set_trim_volume returns 0 on success, else one of these error
	     * codes: FIXME
	     */
	    void (*get_trim_volume_range)(double range[2]);
	    double (*get_trim_volume)(void);
	    int (*set_trim_volume)(double volume);
            void (*get_xyz_balance_range)(double range[2]);
            void (*get_xyz_balance)(double balance[3]);
            int (*set_xyz_balance)(const double balance[3]);
	    /****
	     * vcu.runtime.reverb.get_trim_delay_range
	     * vcu.runtime.reverb.get_trim_delay
	     * vcu.runtime.reverb.set_trim_delay
	     *
	     * Set the overall trim delay for the currently selected reverb
	     * program. The purpose of this is to be able to adjust the delay if
	     * the one stored in the reverb program seems inappropriate. The
	     * setting is stored in a reverb configuration file on disk.
	     *
	     * .set_trim_delay returns 0 on success, else one of these error
	     * codes: FIXME
	     */
	    void (*get_trim_delay_range)(double range[2]);
	    double (*get_trim_delay)(void);
	    int (*set_trim_delay)(double delay);
            /* actually the ones below are not accessible in runtime, but they
               might be in the future */
            void (*get_scale_length_range)(double range[2]);
            double (*get_scale_length)(void);
            int (*set_scale_length)(double scale_length);
            int (*get_response_mode)(void);
            int (*set_response_mode)(int mode);
            int (*get_decorrelate_mode)(void);
            int (*set_decorrelate_mode)(int mode);
            bool_t (*get_force_delay_diff)(void);
            int (*set_force_delay_diff)(bool_t activate);
            bool_t (*get_force_decay)(void);
            int (*set_force_decay)(bool_t activate);
            bool_t (*get_dry_active)(void);
            int (*set_dry_active)(bool_t activate);
            double (*get_dry_volume)(void);
            int (*set_dry_volume)(double volume);
            double (*get_bformat_directivity)(void);
            int (*set_bformat_directivity)(double directivity);
            int (*get_preprocess_properties)(double *tail_volume,
                                             double *length,
                                             double *rt60);
            int (*get_postprocess_properties)(double *tail_volume,
                                              double *length,
                                              double *rt60);
	} reverb;
	struct { /* ambiopole */
	    prgid_t (*get_id)(int idx);
	    int (*set_id)(int idx,
                          prgid_t id);
	} ambiopole;
        struct { /* channel */
            int (*get_mode)(void);
            int (*set_mode)(int idx);
            bool_t (*get_active)(int io,
                                 int type,
                                 int idx);
            bool_t (*get_type_active)(int type);
            bool_t (*get_group_active)(int idx);
        } channel;
    } runtime;
    struct { /* signal */
	bool_t (*get_soft_clip)(void);
	int (*set_soft_clip)(bool_t do_soft_clip);
        int (*get_max_dither_bits)(void);
	bool_t (*get_dither)(void);
	int (*set_dither)(bool_t apply_dither);
        bool_t (*get_watch_clock)(void);
        int (*set_watch_clock)(bool_t watch_clock);
        bool_t (*get_double_precision)(void);
        int (*set_double_precision)(bool_t double_precision);
	void (*get_allowed_output_bits)(int bits[],
					int *n_bits);
	int (*get_output_bits)(void);
	int (*set_output_bits)(int bits);
	void (*get_allowed_rates)(int rates[],
				  int *n_rates);
        void (*get_rates)(int rates[],
			  int *n_rates);
	int (*set_rates)(const int rates[],
			 int n_rates);
	int (*get_suggested_rate)(void);
	int (*set_suggested_rate)(int rate);
    } signal;
    struct { /* channel */
        int (*get_max_count)(int io,
                             int type);
	int (*get_count)(int io,
                         int type);
	int (*set_count)(int type,
                         int n_channels);


        const uint32_t *(*get_phys_map)(int io,
                                        int type,
                                        int idx);

        int (*set_phys_map)(int io,
                            int type,
                            int idx,
                            const uint32_t mask[VCU_MAX_CHANNELS/32]);

        double (*distance_max)(void);
        int (*position_get)(int idx,
                            bool_t *active,
                            double *azimuth,
                            double *elevation,
                            double *distance);
        int (*position_set)(int idx,
                            bool_t activate,
                            double azimuth,
                            double elevation,
                            double distance);
        int (*get_stage_max_count)(void);
	int (*get_stage_count)(void);
	int (*get_stage)(int idx,
                         double *azimuth,
                         double *elevation,
                         double *width,
                         double *height);
	int (*new_stage)(double azimuth,
                         double elevation,
                         double width,
                         double height);
	int (*modify_stage)(int idx,
                            double azimuth,
                            double elevation,
                            double width,
                            double height);
	int (*delete_stage)(int idx);
        bool_t (*get_distance_compensate_active)(void);
        int (*set_distance_compensate_active)(bool_t activate);

	int (*get_group_max_count)(void);
	int (*get_group_count)(void);
	const struct vcu_chgrp *(*get_group)(int idx);
	int (*new_group)(const struct vcu_chgrp *chgrp);
	int (*modify_group)(int idx,
                            const struct vcu_chgrp *chgrp);
	int (*delete_group)(int idx);
        int (*reorder_groups)(const int order[]);

        int (*get_mode_max_count)(void);
        int (*get_mode_count)(void);
        const struct vcu_chmode *(*get_mode)(int idx);
        int (*new_mode)(const struct vcu_chmode *mode);
        int (*modify_mode)(int idx,
                           const struct vcu_chmode *mode);
        int (*delete_mode)(int idx);
        int (*reorder_modes)(const int order[]);
    } channel;
    struct { /* eq */
	int (*get_equaliser_count)(void);
	const struct vcu_eqprg *(*get_equaliser)(int idx);
	int (*new_equaliser)(const struct vcu_eqprg *eqprg);
	int (*modify_equaliser)(int idx,
				const struct vcu_eqprg *eqprg);
	int (*delete_equaliser)(int idx);

#define VCU_FREQ_ISO_3OCT_BANDS 31
	int (*get_equaliser_properties)
            (int idx,
             struct vcu_eqamp amp[VCU_FREQ_ISO_3OCT_BANDS],
             double *peak_amp,
             double *delay);
	double (*get_capacity)(double *max_capacity);
	const struct filterprg *(*get_programs)(int *n_programs);
	int (*load_program)(const struct filterprg *program,
                            bool_t copy_to_cache);
	int (*delete_program)(prgid_t id);

        int (*get_mode_max_count)(void);
        int (*get_mode_count)(void);
        const struct vcu_eqmode *(*get_mode)(int idx);
        int (*new_mode)(const struct vcu_eqmode *mode);
        int (*modify_mode)(int idx,
                           const struct vcu_eqmode *mode);
        int (*delete_mode)(int idx);
        int (*reorder_modes)(const int order[]);
        bool_t (*mode_is_assignable)(int eq_mode_idx,
                                     int channel_mode_idx);
    } eq;
    struct { /* iodelay */
	bool_t (*need_to_run_benchmark)(void);
	bool_t (*has_run_benchmark)(void);
        bool_t (*manually_set)(void);
	int (*run_benchmark)(bool_t test_low_latency,
                             int *limited_by);
        double (*get_benchmark_max_load)(void);
        int (*set_benchmark_max_load)(double load);
        void (*get_benchmark_max_length_range)(double range[2]);
        double (*get_benchmark_max_length)(void);
        int (*set_benchmark_max_length)(double length);
	int (*get_dlp_alternatives)(int sample_rate,
				    double delay[],
				    double max_reverb_length[],
				    int *n_alternatives);
	void (*get_delay_length_pair)(int sample_rate,
				      double *delay,
				      double *max_reverb_length,
				      int *benchmark_index);
	int (*set_delay_length_pair)(int sample_rate,
				     int idx);
        int (*set_delay_length_pair_manual)(int sample_rate,
                                            double delay,
                                            double max_reverb_length);
        double (*get_alignment_delay)(int sample_rate);
        bool_t (*has_truncated_filters)(void);
    } iodelay;
    struct { /* ambiopole */
	bool_t (*get_xtc_active)(int idx);
	int (*set_xtc_active)(int idx,
                              bool_t activate);
	double (*get_capacity)(double *max_capacity);
	const struct filterprg *(*get_programs)(int *n_programs);
	int (*load_program)(const struct filterprg *program,
                            bool_t copy_to_cache);
	int (*delete_program)(prgid_t id);
    } ambiopole;
    struct { /* reverb */
	double (*get_capacity)(double *max_capacity);
	const struct filterprg *(*get_programs)(int *n_programs);
	int (*load_program)(const struct filterprg *program,
                            bool_t copy_to_cache);
        int (*copy_program)(const struct filterprg *program,
                            struct filterprg *destination);
	int (*delete_program)(prgid_t id);
        int (*get_response_mode)(void);
        int (*set_response_mode)(int mode);
        int (*get_decorrelate_mode)(void);
        int (*set_decorrelate_mode)(int mode);
        bool_t (*get_force_delay_diff)(void);
        int (*set_force_delay_diff)(bool_t activate);
        bool_t (*get_force_decay)(void);
        int (*set_force_decay)(bool_t activate);
        bool_t (*get_dry_active)(void);
        int (*set_dry_active)(bool_t activate);
        void (*get_dry_volume_range)(double range[2]);
        double (*get_dry_volume)(void);
        int (*set_dry_volume)(double volume);
        void (*get_bformat_directivity_range)(double range[2]);
        double (*get_bformat_directivity)(void);
        int (*set_bformat_directivity)(double directivity);
        int (*get_source_layout)(void);
        int (*set_source_layout)(int layout);
    } reverb;
    struct { /* ambisonics */
        int (*get_input_format)(void);
        int (*set_input_format)(int format);
    } ambisonics;
    struct { /* cache */
	uint64_t (*get_capacity)(uint64_t *max_capacity);
	const struct filterprg *(*get_programs)(int *n_programs,
						int type);
        const char *(*get_path)(void);
	int (*delete_program)(prgid_t id);
        int (*flush_orphan_resources)(void);
    } cache;
    struct { /* sysinfo */
	int (*get_ram_megabytes)(void);
	double (*get_ram_used)(void);
	int (*get_cpu_megahertz)(void);
	int (*get_cpu_count)(void);
	const char *(*get_cpu_name)(void);
        double (*get_cpu_maxload)(void);
	void (*get_uptime)(int *days,
			   int *hours);
	int (*get_filter_ram_megabytes)(void);
	const char *(*get_version)(void);
        double (*get_temp_cpu)(void);
        double (*get_temp_motherboard)(void);
    } sysinfo;
    struct { /* ctl */
	bool_t (*get_debug)(void);
	void (*set_debug)(bool_t activate);
        bool_t (*get_powersave)(void);
        void (*set_powersave)(bool_t activate);
	bool_t (*need_to_update_hard_settings)(void);
	int (*update_hard_settings)(void);
	void (*schedule_save_config)(void);
	bool_t (*register_read_fd)(int fd,
				   void (*callback)(int fd));
	void (*unregister_read_fd)(int fd);
	int (*register_timer)(void *param,
			      void (*callback)(void *param),
			      struct timeval *timeval);
	void (*unregister_timer)(int id);
	const struct filterprg *(*scan_fprg_dir)(const char path[],
                                                 int type,
                                                 int *n_programs);
        const struct vcu_config *(*scan_config_dir)(const char path[],
                                                    int *n_configs);
        void (*get_load_config_settings)(bool_t *load_user_config,
                                         bool_t *load_revprg_config,
                                         bool_t *load_static_config);
        void (*set_load_config_settings)(bool_t load_user_config,
                                         bool_t load_revprg_config,
                                         bool_t load_static_config);
        int (*load_config)(const struct vcu_config *config);
        int (*save_config)(const char path[],
                           const char name[],
                           bool_t full_path);
	const struct ui_module *(*get_modules)(int *n_modules);
	const char *(*strerror)(int error);
        const char *(*get_log_path)(void);
        const char *(*get_log_file_path)(void);
        void (*flush_logs)(void);
        int (*save_logs)(const char path[]);
        int (*update_software)(const char path[]);
        int (*restart)(bool_t reset_revprg_config,
                       bool_t reset_user_config);
	void (*exit)(int exitcode);
    } ctl;    
};

#endif
