/*
 * (c) Copyright 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _BFIRCONF_H_
#define _BFIRCONF_H_

#include <stdio.h>
#include <inttypes.h>

#include "defs.h"
#include "bfir.h"
#include "bfirdata.h"
#include "almusvcu.h"

#include <bfmod.h>

struct runtime_map {
    uint32_t virt2filt[2][VCU_MAX_CHANNELS][BF_MAXFILTERS/32];
    int rev2filt[VCU_MAX_CHANNELS][3];
    int xtc2filt[VCU_MAX_CHANNELS][4];
    int eq_logi2filt[2][VCU_CH_COUNT][VCU_MAX_CHANNELS];
    uint32_t eq_phys2filt[2][VCU_MAX_CHANNELS][BF_MAXFILTERS/32];

    double out_scales[BF_MAXFILTERS][VCU_MAX_CHANNELS];
    
    int filter_count;
};

struct chmap {
    uint32_t phys2virt[2][VCU_MAX_CHANNELS][VCU_MAX_CHANNELS/32];
    uint32_t logi2virt[2][VCU_CH_COUNT][VCU_MAX_CHANNELS][VCU_MAX_CHANNELS/32];
    int virt2phys[2][VCU_MAX_CHANNELS];
    int virt2logi_type[2][VCU_MAX_CHANNELS];
    int virt2logi_indx[2][VCU_MAX_CHANNELS];
    int virt_count[2];
};

struct brutefir_config {
    int sample_rate;
    int block_length;
    int n_blocks;
    int n_processors;
    bool_t show_progress;
    bool_t apply_dither;
    bool_t powersave;
    bool_t watch_clock;
    bool_t allow_poll_mode;
    bool_t double_precision;
    int cli_pipe[2];
    char convolver_config[PATH_MAX];
    char modules_path[PATH_MAX];
    int channel_count[2][VCU_CH_COUNT];
    int ambisonic_input_format;
    int max_delay[2];

    struct chmap chmap;
    double ampl[2][VCU_MAX_CHANNELS];
    bool_t muted[2][VCU_MAX_CHANNELS];
    int delays[2][VCU_MAX_CHANNELS];
    int device_count[2];
    char device[2][CONF_DEVICES_MAX][PATH_MAX];
    char device_params[2][CONF_DEVICES_MAX][PATH_MAX];
    int sample_format[2];
    int open_channels[2][CONF_DEVICES_MAX];
    int tot_open_channels[2];
    int ext_channel_order[2][CONF_DEVICES_MAX][VCU_MAX_CHANNELS];

    uint32_t eqactive[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS/32];
    int eqcoeffs[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
    bool_t xtalk_cancel[VCU_MAX_CHANNELS/2];
    int xtalkcancel_coeffs[2 * VCU_MAX_CHANNELS];
    int reverb_coeffs[3 * VCU_MAX_CHANNELS];
    int dyneq_count;
    struct bfirdyneq *dyneq[VCU_CH_COUNT * VCU_MAX_CHANNELS];
    double ambisonic_gains[9 * VCU_MAX_CHANNELS];
};

bool_t
bfirconf_config_write(FILE *stream,
                      const struct brutefir_config *config,
                      const struct brutefir_coeff coeffs[],
                      int n_coeffs,
                      struct runtime_map *rmap);

#endif
