/*
 * (c) Copyright 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _BFIRDATA_H_
#define _BFIRDATA_H_

#include "bfir.h"
#include "bfirmem.h"
#include "almusvcu.h"

struct brutefir_coeff {
    int type; /* BFIR_FILTER_TYPE_* */
    int id;
    int index;
    int sub_index;
    int n_blocks;
    int n_segments;
    int segids[BFIR_MAX_SEGMENTS];
    int segoffsets[BFIR_MAX_SEGMENTS];
    int seglengths[BFIR_MAX_SEGMENTS];
};

static inline const char *
bfirdata_typename(int type)
{
    switch (type) {
    case BFIR_FILTER_TYPE_REVERB:
        return "reverb";
    case BFIR_FILTER_TYPE_EQ:
        return "eq";
    case BFIR_FILTER_TYPE_XTC:
        return "xtc";
    case BFIR_FILTER_TYPE_UHJ:
        return "uhj";
    }
    return "##error##";
}

int
bfirdata_new_dummy_filter(int n_blocks,
                          int block_size);

const struct brutefir_coeff *
bfirdata_get_dummy_coeffs(int *n_coeffs);

int
bfirdata_filter_count(int type);

int
bfirdata_longest_filter_blocks(int sample_rate);
    
int
bfirdata_coeff_index(int type,
                     int id,
                     int sub_index);

int
bfirdata_new_filters(int type,
                     struct bfirfilter filters[],
                     const int lengths[]);

void
bfirdata_delete_filters(int id);

int
bfirdata_shorten_filters(int id,
                         const int lengths[]);

const struct brutefir_coeff *
bfirdata_get_coeffs(int *n_coeffs,
                    int sample_rate);

double
bfirdata_filter_capacity(int type,
                         double *max_capacity);

void
bfirdata_reset(void);
    
void
bfirdata_settings(int n_reverb_input_channels,
                  int n_reverb_output_channels,
                  int n_rates,
                  const int rates[],
                  const int block_sizes[],
                  int realsize);

bool_t
bfirdata_init(int alloc_bytes);

#endif
