/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include "apui_defs.h"
#include "path.h"

char *
path_tilde_expansion(const char path[])
{
    static char real_path[PATH_MAX];
    struct passwd *pw;
    char *homedir = NULL, *p;
    int pos, slashpos = -1;
    
    if (path[0] != '~') {
	strncpy(real_path, path, PATH_MAX);
	real_path[PATH_MAX-1] = '\0';
	return real_path;
    }

    if ((p = strchr(path, PATH_SEPARATOR_CHR)) != NULL) {
	slashpos = p - path;
    }
    
    if (path[1] == '\0' || slashpos == 1) {
	if ((homedir = getenv("HOME")) == NULL) {
	    if ((pw = getpwuid(getuid())) != NULL) {
		homedir = pw->pw_dir;
	    }
	}
    } else {
	strncpy(real_path, &path[1], PATH_MAX);
	real_path[PATH_MAX-1] = '\0';
	if (slashpos != -1) {
	    real_path[slashpos-1] = '\0';
	}
	if ((pw = getpwnam(real_path)) != NULL) {
	    homedir = pw->pw_dir;
	}
    }
    if (homedir == NULL) {
	strncpy(real_path, path, PATH_MAX);
	real_path[PATH_MAX-1] = '\0';
	return real_path;
    }
    
    real_path[PATH_MAX-1] = '\0';
    strncpy(real_path, homedir, PATH_MAX - 2);
    pos = strlen(homedir);
    if (homedir[0] == '\0' ||
	homedir[strlen(homedir)-1] != PATH_SEPARATOR_CHR)
    {
	strcat(real_path, PATH_SEPARATOR_STR);
	pos += 1;
    }
    if (slashpos != -1) {
	strncpy(&real_path[pos], &path[slashpos+1], PATH_MAX - pos - 1);
    } else {
	strncpy(&real_path[pos], &path[1], PATH_MAX - pos - 1);
    }
    real_path[PATH_MAX-1] = '\0';
    
    return real_path;
}

char *
path_make_absolute(char _path[PATH_MAX],
                   bool_t end_with_slash,
                   bool_t modify_source)
{
    static char new_path[PATH_MAX];
    char path[PATH_MAX];
    int n;

    strcpy(path, path_tilde_expansion(_path));
    n = 0;
    if (path[0] != PATH_SEPARATOR_CHR) {
        /* extend to absolute path */
        getcwd(new_path, PATH_MAX);
        n = strlen(new_path);
        if (new_path[n - 1] != PATH_SEPARATOR_CHR) {
            new_path[n] = PATH_SEPARATOR_CHR;
            new_path[n + 1] = '\0';
            n += 1;
        }
    }
    strncpy(&new_path[n], path, PATH_MAX - n);
    new_path[PATH_MAX - 2] = '\0';
    n = strlen(new_path);
    if (end_with_slash && new_path[n - 1] != PATH_SEPARATOR_CHR) {
        new_path[n] = PATH_SEPARATOR_CHR;
        new_path[n + 1] = '\0';
    }
    if (modify_source) {
        strcpy(_path, new_path);
        return _path;
    }
    return new_path;
}

bool_t
path_create(const char _path[])
{
    char *path, *p, *p1;

    if (_path == NULL || _path[0] == '\0') {
        errno = EINVAL;
        return false;
    }
    path = strdup(_path);
    p = path;
    if (p[0] == PATH_SEPARATOR_CHR) {
        p1 = strchr(&p[1], PATH_SEPARATOR_CHR);
    } else {
        p1 = strchr(p, PATH_SEPARATOR_CHR);
    }
    while (p1 != NULL) {
        *p1 = '\0';
        if (mkdir(path, 0777) == -1 && errno != EEXIST) {
            return false;
        }
        *p1 = PATH_SEPARATOR_CHR;
        p = &p1[1];
        p1 = strchr(p, PATH_SEPARATOR_CHR);
    }
    if (p[0] != '\0') {
        if (mkdir(path, 0777) == -1 && errno != EEXIST) {
            return false;
        }
    }
    free(path);
    return true;
}
