/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _REVERB_H_
#define _REVERB_H_

#include "bfir.h"
#include "almusvcu.h"
#include "filterprg.h"

#define REVERB_BFORMAT_DIRECTIVITY_DEFAULT 1.4
#define REVERB_BFORMAT_DIRECTIVITY_MIN 0.0
#define REVERB_BFORMAT_DIRECTIVITY_MAX 2.0

/* must be run first */
int
reverb_settings(double bformat_directivity_factor,
                double early_response_length,
                double direct_sound_length,
                double dry_level,
                int response_mode,
                int decorrelate_mode,
                bool_t force_delay_diff,
                bool_t force_decay,
                int source_layout,
                int n_stages,
                const double azimuth[],
                const double elevation[],
                const double width[],
                const double height[]);

double
reverb_postprocess_maxlength(double original_length,
                             double predelay);

int
reverb_apply_balance(const double xyz_balance[3],
                     int n_reverb_speakers,
                     const double azimuth[],
                     const double elevation[],
                     double volume[]);

bool_t
reverb_test_receiver_directions(const double azimuth[],
                                const double elevation[],
                                int n_receivers);

int
reverb_preprocess(const struct filterprg *fprg,
                  int realsize,
                  double target_length,
                  int n_reverb_speakers,
                  const double azimuth[],
                  const double elevation[],
                  struct bfirfilter target[],
                  double *orig_rt60,
                  double *orig_tail_volume,
                  double *new_rt60,
                  double *new_tail_volume,
                  double *new_length,
                  bool_t *truncated,
                  void (*progress_callback)(int channel_index,
                                            int n_channels,
                                            int from_rate,
                                            int to_rate,
                                            bool_t deriving));

#endif
